;;;	      Structure and Interpretation of Computer Programs
;;;			       Second Edition 
;;;                          Sample Problem Set 
;;;
;;;			    Code file GAME.SCM
;;;
;;; This file contains class definitions and some general code


;;; ---------------------------------------------------------------------------
;;; Simple object system with inheritance

(define (ask object message . args)  
  (let ((method (get-method object message)))
    (if (method? method)
	(apply method (cons object args))
	(error "No method" message (cadr method)))))

(define (get-method object message)
  (object message))

(define (no-method name)
  (list 'no-method name))

(define (method? x)
  (not (no-method? x)))

(define (no-method? x)
  (if (pair? x)
      (eq? (car x) 'no-method)
      #f))

;;; ----------------------------------------------------------------------------
;;; Persons, places, and things will all be kinds of named objects

;; short-desc, long-desc, and related accessors added
;; For places, short-desc shown on entry and long-desc on 'look
;; For mobile-objects, short-desc shown in listings, long-desc on 'look-at
;; This class may be used for decorations - they don't need to move,
;;   be owned, or even know where they are, just have descriptions.
;; Only persons and places have names now, so named-object is a misnomer
(define (make-named-object keywords short-desc)
  (let ((long-desc 'blank))
    (lambda (message) 
      (cond ((eq? message 'short-desc) (lambda (self) short-desc))
            ((eq? message 'keywords) (lambda (self) keywords))
            
            ((eq? message 'add-long-desc)
             (lambda (self longstring)
               (set! long-desc longstring)
               #t))
          
            ((eq? message 'long-desc) 
             (lambda (self) long-desc ))
          
            (else (no-method short-desc))))))

;;; Persons and things are mobile since their places can change
;;   and keywords is a list of keywords used to reference the
;;   object - Keywords should be prominent in short-desc!
(define (make-mobile-object keywords short-desc location)
  (let ((named-obj (make-named-object keywords short-desc)))
    (lambda (message)
      (cond ((eq? message 'place)    (lambda (self) location))
            
	    ((eq? message 'install)
	     (lambda (self)
	       (ask location 'add-thing self)))	; Synchonize thing and place
            
	    ;; Following method should never be called by the user...
	    ;;  it is a system-internal method.
	    ;; See CHANGE-PLACE instead
	    ((eq? message 'set-place)
	     (lambda (self new-place)
	       (set! location new-place)
	       'place-set))
	    (else (get-method named-obj message))))))

(define (make&install-mobile-object keywords short-desc place)
  (let ((mobile-obj (make-mobile-object keywords short-desc place)))
    (ask mobile-obj 'install)
    mobile-obj))

;;; A thing is something that can be owned
;; Thinking of replacing owner with setting place as holder
;;   and replacing 'ownable? with 'takeable?
;; These will be generic things as opposed to unique things
(define (make-thing keywords short-desc birthplace)
  (let ((owner     'nobody)
	(mobile-obj (make-mobile-object keywords short-desc birthplace)))
    (lambda (message)
      (cond ((eq? message 'owner)    (lambda (self) owner))
	    ((eq? message 'ownable?) (lambda (self) #t))
            
	    ((eq? message 'owned?)
	     (lambda (self)
	       (not (eq? owner 'nobody))))
            
	    ;; Following method should never be called by the user...
	    ;;  it is a system-internal method.
	    ;; See TAKE and LOSE instead.
            ((eq? message 'set-owner)
	     (lambda (self new-owner)
	       (set! owner new-owner)
	       'owner-set))
	    (else (get-method mobile-obj message))))))

(define (make&install-thing keywords short-desc birthplace)	
  (let ((thing  (make-thing keywords short-desc birthplace)))
    (ask thing 'install)
    thing))

;;; Implementation of places
;; Persons may only access things via their currect place using keywords
;; keywords are (until I think of a use for them) unused in places
;;   I might just dissociate places from named-objects anyway
(define (make-place name short-desc)
  (let ((neighbor-map '())		
	(things       '())
	(named-obj (make-named-object '() short-desc)))
    (lambda (message)
      (cond ((eq? message 'name) (lambda (self) name))
            ((eq? message 'gender) (lambda (self) gender))
            ((eq? message 'place?) (lambda (self) #t))
	    ((eq? message 'neighbors) (lambda (self) (map cdr neighbor-map)))
	    ((eq? message 'exits) (lambda (self) (map car neighbor-map)))
            ((eq? message 'things) (lambda (self) things))
            
	    ((eq? message 'neighbor-towards)
	     (lambda (self direction)
	       (let ((places (assq direction neighbor-map)))
		 (if places
		     (cdr places)
		     #f))))
            
            ((eq? message 'add-neighbor)
             (lambda (self direction new-neighbor)
               (cond ((assq direction neighbor-map)
                      (display-message (list "Direction already assigned" direction name))
		      #f)
                     (else
                      (set! neighbor-map (cons (cons direction new-neighbor) neighbor-map))
		      #t))))
            
	    ((eq? message 'accept-person?)
	     (lambda (self person) #t))
            
	    ;; Following two methods should never be called by the user...
	    ;;  they are system-internal methods. See CHANGE-PLACE instead.
            ((eq? message 'add-thing)
             (lambda (self new-thing)
               (cond ((memq new-thing things)
                      (display-message (list (ask new-thing 'name) "is already at" name))
		      #f)
                     (else (set! things (cons new-thing things))
			   #t))))
            
            ((eq? message 'del-thing)
             (lambda (self thing)
               (cond ((not (memq thing things))
                      (display-message (list (ask thing 'name) "is not at" name))
		      #f)
                     (else (set! things (delq thing things))	;; DELQ defined
			   #t))))                               ;; below

            (else (get-method named-obj message))))))

;;; ----------------------------------------------------------------------------
;;; Implementation of people

;; Keywords for persons should be the symbol version of their name,
;;   so a person with name "Bob" should have keywords '(bob)
;;   gender must be 'male, 'female, or 'other
(define (make-person name keywords short-desc birthplace threshold gender maxhp)
  (let ((possessions '())
        (hp maxhp)
	(mobile-obj  (make-mobile-object keywords short-desc birthplace)))
    (lambda (message)
      (cond ((eq? message 'name)        (lambda (self) name))
            ((eq? message 'person?)     (lambda (self) #t))
            ((eq? message 'possessions) (lambda (self) possessions))
            ((eq? message 'get-hp)      (lambda (self) hp))
            ((eq? message 'get-maxhp)   (lambda (self) maxhp))
            
            ;; new message to aid multiplay (no chance of it ever getting going, but...)
            ((eq? message 'show-message)
             (lambda (self message)
               (if (eq? self player)
                   (display-message message))))
            
            ;; for taking damage or healing
            ((eq? message 'damage)
             (lambda (self amount source)    ;; add type later
               (set! hp (- hp amount))
               (display-local self 
                              (list source "hit you for" amount "damage!")
                              (list source "hit" name "for" amount "damage!"))
               (cond ((< hp 0)
                      (display-local self
                                     (list "You are dead!")
                                     (list name "is dead!")))
                     ((= hp 0)
                      (display-local self
                                     (list "You are critically wounded!")
                                     (list name "is critically wounded!")))
                     (else 'nothing-special))))
            
            ;; new message - reflects current state
            ((eq? message 'condition)
             (lambda (self)
               (cond ((= hp maxhp) (list name "is in perfect condition"))
                     ((= hp 0) (list name "is critically wounded."))
                     ((< hp 0) (list name "is dead."))
                     (else (list name "is somewhat hurt.")))))
                 ;; list of special statuses here somehow
            
            ;; if dead (hp<0), say so
            ((eq? message 'dead?)
             (lambda (self)  (< hp 0)))
            
            ;; was 'list-possesions
	    ((eq? message 'inv)
	     (lambda (self)
	       (display-local self
                              (cons "You have" (pretty-object-list possessions))
                              '() )
               'listed))
            
	    ((eq? message 'say)
	     (lambda (self list-of-stuff)
               (if (null? list-of-stuff)
                   (display-local self
                                  '("Say what?")
                                  '() )
                   (display-local self
                                  (cons "You say" list-of-stuff)
                                  (cons name (cons "says" list-of-stuff))))))
            
            ;; emote is incredibly fun and useful when there's anyone around to see it
            ((eq? message 'emote)
             (lambda (self whatever)
               (display-local self
                              (cons name whatever)
                              (cons name whatever))))
            
            ;; method have-fit removed
            ;; method look moved to commands file

            ;; no longer allows taking from others
	    ((eq? message 'take)
	     (lambda (self keywords)
	       (cond ((procedure? keywords)  ;;internal use
                      (cond ((memq keywords possessions)
                             (display-local self (list "You already have that") '() ))
                            ((not (is-a keywords 'ownable?))
                             (display-local self (list "You can't take that") '() ))
                            ((ask keywords 'owned?)       ;; ditch this junk
                             (let ((owner (ask (ask keywords 'owner) 'short-desc)))
                             (display-local self (list owner "has it") '() )))
                            (else
                             (ask keywords 'set-owner self)
                             (set! possessions (cons keywords possessions))
                             (display-local self (list "You take" (ask keywords 'short-desc))
                                                 (list name "takes" (ask keywords 'short-desc))))))
                     ((eq? (car keywords) 'all)
                      (for-each (lambda (object) 
                                  (if (is-a object 'ownable?) (ask self 'take object)))
                                (ask (ask self 'place) 'things) )
                      #t)
                     (else
                      (let ((thing (keyword-lookup (ask (ask self 'place) 'things) keywords)))
                        (cond ((null? thing)
                               (if (not (null? (keyword-lookup possessions keywords)))
                                   (display-local self (list "You already have it") '() )
                                   (display-local self (list "You don't see that" '() )))
                               #f)
                              (else
                               (ask self 'take (car thing))
                               #t)))))))
            
            ;; was 'lose
	    ((eq? message 'drop)
	     (lambda (self keywords)
	       (cond ((procedure? keywords)     ;; internal use
                      (let ((thing keywords))
                        (cond ((eq? self (ask thing 'owner))
                               (set! possessions (delq thing possessions)) ;; DELQ
                               (ask thing 'set-owner 'nobody)              ;; defined
                               (display-message                            ;; below
                                (list name "drops" (ask thing 'short-desc)))
                               #t)
                              (else
                               (display-message (list name "does not have"
                                                      (ask thing 'short-desc)))
                               #f))))
                     ((eq? (car keywords) 'all)
                      (for-each (lambda (thing) (ask self 'drop thing)) possessions)
                      #t)
                     (else
                      (let ((thing (keyword-lookup possessions keywords)))
                        (cond ((null? thing)
                               (display-message
                                (list name "doesn't have that"))
                               #f)
                              (else (ask self 'drop (car thing))
                                    #t)))))))
            
            ;; 0 threshold now a special case
	    ((eq? message 'move)
	     (lambda (self)
	       (cond ((= threshold 0) #f)
                     ((= (random threshold) 0)
		      (ask self 'act)
		      #t))))
            
	    ((eq? message 'act)
	     (lambda (self)
	       (let ((direction (random-exit (ask self 'place))))
		 (if direction
		     (ask self 'go (list direction))
		     #f))))		; All dressed up and no place to go

	    ((eq? message 'move-to)
	     (lambda (self new-place)
	       (let ((old-place (ask self 'place)))
		 (cond ((eq? new-place old-place)
			(display-local self
                                       (list "You're already there")
                                       '() ))
		       ((ask new-place 'accept-person? self)
			(change-place self new-place)
			(for-each (lambda (p) (change-place p new-place)) possessions)
                        (display-local self
                                       (list (ask new-place 'name) "-" (ask new-place 'short-desc))
                                       (list name "enters the room"))
                        (display-local self
                                       (cons "Visible exits: " (ask new-place 'exits))
                                       '())
                        (let ((other-things (delq self (ask (ask self 'place) 'things))))
                          (if (not (null? other-things))
                              (display-local self
                                             (cons "   You see" (pretty-object-list other-things))
                                             '() ))))
		       (else
			(display-local self
                                       (list "You can't move to" (ask new-place 'name))
                                       '() ))))))
            
	    ((eq? message 'go)
	     (lambda (self direction)
	       (let ((old-place (ask self 'place))
                     (direction (car direction)))      ;; since it's incoming as a list
		 (let ((new-place (ask old-place 'neighbor-towards direction)))
		   (cond (new-place
                          (display-local self
                                         '()
                                         (list name "leaves" direction))
			  (ask self 'move-to new-place))
			 (else
			  (display-local self
                                         (list "You cannot go" direction "from" (ask old-place 'name))
                                         '())
			  #f))))))
            
            ;; method exits moved to commands
            ;; method look-at moved to commands
            
	    ((eq? message 'install)
	     (lambda (self)
	       (add-to-clock-list self)
	       ((get-method mobile-obj 'install) self)))
            
	    (else (get-method mobile-obj message))))))
  
(define (make&install-person name keywords short-desc birthplace threshold gender maxhp)
  (let ((person (make-person name keywords short-desc birthplace threshold gender maxhp)))
    (ask person 'install)
    person))

;;; A troll is a kind of person (but not a kind person!)

(define (make-troll name keywords short-desc birthplace threshold maxhp)
  (let ((person (make-person name keywords short-desc birthplace threshold 'other maxhp)))
    (lambda (message)
      (cond ((eq? message 'act)
	     (lambda (self)
	       (let ((others (other-people-at-place self (ask self 'place))))
		 (if (not (null? others))
		     (ask self 'eat-person (pick-random others))
		     ((get-method person 'act) self)))))
            
            ;; modified to disallow eating self
	    ((eq? message 'eat-person)
	     (lambda (self person)
               (cond ((not (eq? person self))
                      (ask self 'say
                           (list "Growl...." (ask self 'name) "gonna eatcha,"
                                 (ask person 'name)))
                      (ask person 'damage 61 name)
                      (cond ((ask person 'dead?)
                             (go-to-heaven person self)
                             (ask self 'say
                                  (list "Chomp chomp." (ask person 'name)
                                        "taste yummy!"))
                             '*burp*)
                            (else
                             (ask self 'say
                                  (list "Graahgh!" (ask person 'name) "gonna be troll food!")))))
                     (else (ask self 'say (list (ask self 'name) "no can eat" (ask self 'name))) 'nyaaah))))
            
	    (else (get-method person message))))))

(define (make&install-troll name keywords short-desc birthplace threshold maxhp)
  (let ((troll  (make-troll name keywords short-desc birthplace threshold maxhp)))
    (ask troll 'install)
    troll))


(define (go-to-heaven person killer)
  (for-each (lambda (item) (ask person 'drop item))
	    (ask person 'possessions))
  (ask person 'say '("Dulce et decorum est"))
  (ask person 'say '("pro computatore mori!"))
  (ask person 'move-to heaven)
  (remove-from-clock-list person)
  (display-message (list (ask person 'name) "was killed by" (ask killer 'name)))
  'game-over-for-you-dude)

;; (define heaven (make-place 'heaven))		; The point of no return

;;; --------------------------------------------------------------------------
;;; Clock routines

(define *clock-list* '())
(define *the-time* 0)

(define (initialize-clock-list)
  (set! *clock-list* '())
  'initialized)

(define (add-to-clock-list person)
  (set! *clock-list* (cons person *clock-list*))
  'added)

(define (remove-from-clock-list person)
  (set! *clock-list* (delq person *clock-list*))  ;; DELQ defined below
  'removed)

(define (clock)
;;  (display "---Tick---")
;;  (newline)
  (set! *the-time* (+ *the-time* 1))
  (for-each (lambda (person) (ask person 'move))
	    *clock-list*)
  'tick-tock)
	     

(define (current-time)
  *the-time*)

(define (run-clock n)
  (cond ((zero? n) 'done)
	(else (clock)
	      (run-clock (- n 1)))))

;;; --------------------------------------------------------------------------
;;; Miscellaneous procedures

(define (is-a object property)
  (let ((method (get-method object property)))
    (if (method? method)
	(ask object property)
	#f)))

(define (change-place mobile-object new-place)	; Since this bridges the gap
  (let ((old-place (ask mobile-object 'place))) ; between MOBILE-OBJECT and
    (ask mobile-object 'set-place new-place)	; PLACE, it is best it not
    (ask old-place 'del-thing mobile-object))	; be internal to either one.
  (ask new-place 'add-thing mobile-object)
  'place-changed)

(define (other-people-at-place person place)
  (filter (lambda (object)
	    (if (not (eq? object person))
		(is-a object 'person?)
		#f))
	  (ask place 'things)))

;;greet-people removed since it's pointless

;; new function - creates comma deliniated list of objects (for output)
(define (pretty-object-list list-of-stuff)
  (cond ((null? list-of-stuff) (list "nothing") )
        ((null? (cdr list-of-stuff)) (cons (ask (car list-of-stuff) 'short-desc) '() ) )
        (else (cons
               (string-append (ask (car list-of-stuff) 'short-desc) ",")
               (pretty-object-list (cdr list-of-stuff))))))

;; new function - finds and returns object in list matching keywords
(define (keyword-lookup objects keywords)
  (cond ((null? objects) '() )           ;; no match found
        ((null? keywords) objects)       ;; return any matches
        ((keyword-lookup 
          (filter (lambda (obj)
                    (memq (car keywords) (ask obj 'keywords)))
                  objects)
          (cdr keywords)))))

(define (display-message list-of-stuff)
  (for-each (lambda (s) (display s) (display " "))
	    list-of-stuff)
  (newline)
  'message-displayed)

(define (random-exit place)
  (pick-random (ask place 'exits)))

(define (filter predicate lst)
  (cond ((null? lst) '())
	((predicate (car lst))
	 (cons (car lst) (filter predicate (cdr lst))))
	(else (filter predicate (cdr lst)))))

(define (pick-random lst)
  (if (null? lst)
      #f
      (list-ref lst (random (length lst)))))  ;; See manual for LIST-REF

(define (delq item lst)
  (cond ((null? lst) '())
	((eq? item (car lst)) (delq item (cdr lst)))
	(else (cons (car lst) (delq item (cdr lst))))))

;;; -------------------------------------------------------------------
;;; Other interesting procedures for PS6, Spring 95

(define (make&install-sd-card name short-desc birthplace id)
  (let ((card (make-sd-card name short-desc birthplace id)))
    (ask card 'install)
    card))

(define (make-sd-card name short-desc birthplace idnumber)
  (let ((id idnumber)
        (thing (make-thing name short-desc birthplace)))
    (lambda (message)
      (cond ((eq? message 'sd-card?) (lambda (self) #t))
            ((eq? message 'id) (lambda (self) id))
            (else (get-method thing message))))))

(define (copy-sd-card card)
  (let ((name   (symbol-append 'copy-of- (ask card 'name)))
	(place  (ask card 'place))
	(id     (ask card 'id)))
    (make&install-sd-card name place id)))

(define (show thing)
  (define (global-environment? frame)
    (environment->package frame))
  (define (pp-binding name value width)
    (let ((value* (with-string-output-port
		   (lambda (port)
		     (if (pair? value)
			 (pretty-print value port #F (+ width 2))
			 (display value port))))))
      (newline)
      (display name)
      (display ": ")
      (display (make-string (- width (string-length name)) #\Space))
      (if (pair? value)
	  (display (substring value* (+ width 2) (string-length value*)))
	  (display value*))))
  (define (show-frame frame)
    (if (global-environment? frame)
	(display "\nGlobal Environment")
	(let* ((bindings (environment-bindings frame))
	       (parent   (environment-parent frame))
	       (names    (cons "Parent frame"
			       (map symbol->string (map car bindings))))
	       (values   (cons (if (global-environment? parent)
				   'global-environment
				   parent)
			       (map cadr bindings)))
	       (width    (reduce max 0 (map string-length names))))
	  (for-each (lambda (n v) (pp-binding n v width))
	    names values))))
  (define (show-procedure proc)
    (fluid-let ((*unparser-list-depth-limit* 4)
		(*unparser-list-breadth-limit* 4))
      (newline)
      (display "Frame:")
      (newline)
      (display "  ")
      (if (global-environment? (procedure-environment proc))
	  (display "Global Environment")
	  (display (procedure-environment proc)))
      (newline)
      (display "Body:")
      (newline)
      (pretty-print (procedure-lambda proc) (current-output-port) #T 2)))

  (define (print-nicely thing)
    (newline)
    (display thing)
    (cond ((not thing) ;;formerly false?
	   'uninteresting)
	  ((environment? thing)
	   (show-frame thing))
	  ((compound-procedure? thing)
	   (show-procedure thing))
	  (else 'uninteresting)))

  (print-nicely
   (or (if (integer? thing);;formerly exact-integer?
	   (object-unhash thing)
	   thing)
       thing)))